/*
** Module   : Configuration.java
** Abstract : Master configuration object
**
** Copyright (c) 2004-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20041217   @19157 Created initial version.
** 002 ECF 20050120   @19359 Added support for individual parameters with
**                           variable substitution. New public methods:
**                           - getParameter (two flavors)
**                           - putParameter
** 003 GES 20050223   @19931 Added a public utility method to normalize
**                           a passed in filename to the project root so
**                           that tools can refer to the same file from
**                           different current directories without having
**                           to use absolute filenames.
** 004 ECF 20050225   @19963 Modified toFile method to canonicalize file.
** 005 GES 20050617   @21507 Moved sweep() to StringHelper.
** 006 GES 20070411   @32924 Forced a default value of "." for P2J_HOME
**                           since this is essentially the only value
**                           we ever use anyway. This is just safer and
**                           easier than forcing everyone to always
**                           specify a value.
** 007 GES 20070717   @34580 Exposed the ${SYMBOL} style parameter
**                           substitution as an external service. Any of
**                           the parameters in the parmMap will be matched
**                           and replaced as is normally done during
**                           putParameter(). This allows other config
**                           files to act like and reference the main
**                           system config.
** 008 GES 20090424   @41927 Import change.
** 009 LMR 20110105          Replaced Apache Digester usage with direct SAX
**                           XML processing using and the tools in 
**                           GenericSAXHandler and Mapper; performed other
**                           refactoring: 
**                             -renamed load() to getSchemaConfig()
**                             -removed the 'key' parameter and throws clause
**                              of getSchemaConfig()
**                             -inlined code in initialize() in the static
**                              initializer block and removed initialize()
**                             -added main() method to test the parsing of
**                              the XML file received in args[0]
** 010 GES 20110712          Added a helper to return back parameter values
**                           already converted to the proper type.
** 011 ECF 20130125          Added schema metadata configuration. Added generics.
*/

package com.goldencode.p2j.cfg;

import java.io.*;
import java.util.*;
import javax.xml.parsers.*;
import org.xml.sax.*;
import com.goldencode.util.*;
import com.goldencode.p2j.schema.*;
import com.goldencode.p2j.schema.SchemaConfig.*;
import com.goldencode.p2j.schema.SchemaConfig.NsConfig.*;

/**
 * Master configuration object which loads subsystem configurations. The
 * main configuration and all subsystem configurations are stored as XML
 * documents, which are expected to be found in a <code>cfg</code>
 * subdirectory of the P2J home path. The home path is defined by the
 * required <code>P2J_HOME</code> system property.
 * <p>
 * Subsystem configurations are java objects returned by the {@link 
 * #getSchemaConfig()} method. They are created from XML configuration 
 * information by a SAX parser using the <code>GenericSAXHandler</code> and 
 * <code>Mapper</code> classes in pacakge <code>com.goldencode.p2j.util
 * </code>. XML configuration data is stored in a file named <code>p2j.cfg.xml
 * </code>.
 * <p>
 * This class also provides services to get and store individual parameter
 * entries which are read from the main configuration file upon initialization
 * (from cfg/global/parameter XML elements). 
 * <p>Variable substitution is supported in these parameters, in the form 
 * <code>${key_name}</code>, where <code>key_name</code> is the name of a 
 * parameter previously defined by this mechanism. The replacement value 
 * applied to the substitution variable is the value of the parameter 
 * identified by <code>key_name</code>. For example,
 * <pre>
 *    &lt;cfg&gt;
 *      &lt;global&gt;
 *         &lt;parameter name="user.home" value="/home/user" /&gt;
 *         &lt;parameter name="subdir"    value="${user.home}/mysubdir" /&gt;
 *      &lt;/global&gt;
 *      ...
 *    &lt;/cfg&gt;
 * </pre>
 * results in the following parameter definitions after substitution:
 * <pre>
 *    user.home = /home/user
 *    subdir    = /home/user/mysubdir
 * </pre>
 * <strong>Note:</strong> the following restrictions apply to this
 * mechanism:
 * <ul>
 * <li><code>P2J_HOME</code> is a reserved parameter key. It may not be 
 *     defined in the configuration file, but it is available for use as a
 *     substitution variable.</li>
 * <li>The variable name <code>key_name</code> must not include any
 *     characters which have special syntax meaning in regular expressions.
 *     Regular expressions are used internally for the replacement; special
 *     characters in <code>key_name</code> are not explicitly escaped,
 *     which could cause unexpected results.</li>
 * </ul>
 */
public final class Configuration
{
   /** Constant denoting the <code>&lt;cfg&gt;</code> XML tag */
   public static final String TAG_CFG = "cfg";

   /** Constant denoting the <code>&lt;global&gt;</code> XML tag */
   public static final String TAG_GLOBAL = "global";

   /** Constant denoting the <code>&lt;parameter&gt;</code> XML tag */
   public static final String TAG_PARAMETER = "parameter";
   
   /** Constant denoting the <code>&lt;schema&gt;</code> XML tag */
   public static final String TAG_SCHEMA = "schema";
   
   /** Constant denoting the <code>&lt;namespace&gt;</code> XML tag */
   public static final String TAG_NAMESPACE = "namespace";
   
   /** Constant denoting the <code>&lt;dialect-specific&gt;</code> XML tag */
   public static final String TAG_DIALECTSPECIFIC = "dialect-specific";
   
   /** Constant denoting the <code>&lt;metadata&gt;</code> XML tag */
   public static final String TAG_METADATA = "metadata";
   
   /** Constant denoting the <code>&lt;table&gt;</code> XML tag */
   public static final String TAG_TABLE = "table";
   
   /** System property name for P2J home path */
   private static final String PROP_HOME_DIR = "P2J_HOME";
   
   /** Configuration subdirectory name */
   private static final String CFG_DIR = "cfg";
   
   /** Configuration file name extension */
   private static final String CFG_EXTENSION = ".cfg.xml";
   
   /** Default configuration filename */
   private static final String DEF_CFG_FILE = "p2j.cfg.xml";
   
   /** Map of parameters */
   private static final Map<String, String> parmMap = new LinkedHashMap<String, String>();
   
   /** Factory providing the SAX parser instance that will be used to parse the config file. */
   private static SAXParserFactory saxParserFactory;
   
   /** The parser instance. */
   private static SAXParser saxParser;
   
   /**
    * The generic SAX handler that will be used by the parser after it's 
    * configured to work with P2J config files. 
    */
   private static GenericSAXHandler saxHandler;
   
   // Initialize this class. 
   static
   {
      try
      {
         // Always start with P2J_HOME parameter; this is reserved.
         parmMap.put(PROP_HOME_DIR, home());

         File cfgFile = getConfigFile(getConfigDirectory(), DEF_CFG_FILE);
         saxParserFactory = SAXParserFactory.newInstance();
         saxParser = saxParserFactory.newSAXParser();
         saxHandler = new GenericSAXHandler();
            
         // Prepare the generic handler.
         setupHandler(saxHandler);

         // Parse the file.
         loadXml(cfgFile);

         // Map parameters by their names, which should be unique.
         List<?> params = saxHandler.getGlobalObjects(TAG_PARAMETER);
         for (Object o : params)
         {
            Parameter p = (Parameter) o;
            putParameter(p.getName(), p.getValue());
         }  
      }
      catch (Exception exc)
      {
         // We cannot throw a checked exception from <clinit>, so if we hit a
         // fatal problem, wrapper the checked exception in a RuntimeException
         // and throw it. This results in an ErrorInInitializerError on this
         // thread. It is arguable whether this is any better than an abend.
         // TODO: come up with a better solution here...
         throw new RuntimeException(new ConfigurationException(
            "Error loading " + DEF_CFG_FILE + " configuration", exc));
      }
   }
   
   /**
    * Private default constructor to prevent creation of instances of this
    * class, since all methods are static.
    */
   private Configuration()
   {
   }

   /**
    * Return a schema configuration object. The configuration data 
    * itself is in the default XML configuration file <code>p2j.cfg.xml</code>
    * that was parsed when this class was loaded.
    *
    * @return  The configuration object loaded from XML.
    */
   public static SchemaConfig getSchemaConfig()
   {
      return (SchemaConfig) saxHandler.getGlobalObjects(TAG_SCHEMA).get(0);
   }
   
   /**
    * Get the configuration parameter with the specified key. This may have
    * been set in the main configuration file, or added later using the
    * {@link #putParameter} method.
    *
    * @param   key
    *          Name of parameter to retrieve.
    *
    * @return  The value of the specified parameter, or <code>null</code> if
    *          <code>key</code> is an invalid identifier.
    */
   public static String getParameter(String key)
   {
      return getParameter(key, null);
   }
   
   /**
    * Get the configuration parameter with the specified key. This may have
    * been set in the main configuration file, or added later using the
    * {@link #putParameter} method.
    *
    * @param   key
    *          Name of parameter to retrieve.
    * @param   defaultValue
    *          Value to be returned if <code>key</code> cannot be found in
    *          the map of parameters. May be <code>null</code>.
    *
    * @return  The value of the specified parameter, or
    *          <code>defaultValue</code> if <code>key</code> is not found.
    */
   public static String getParameter(String key, String defaultValue)
   {
      String value = parmMap.get(key);
      
      return (value != null ? value : defaultValue);
   }
   
   /**
    * Get the configuration parameter with the specified key as a boolean value.
    * This may have been set in the main configuration file, or added later
    * using the {@link #putParameter} method.
    *
    * @param   key
    *          Name of parameter to retrieve.
    * @param   defaultValue
    *          Value to be returned if <code>key</code> cannot be found in
    *          the map of parameters.
    *
    * @return  The value of the specified parameter, or
    *          <code>defaultValue</code> if <code>key</code> is not found.
    */
   public static boolean getParameter(String key, boolean defaultValue)
   {
      String value = parmMap.get(key);
      
      return (value != null ? "true".equals(value) : defaultValue);
   }
   
   /**
    * Associate a configuration parameter value with <code>key</code>.
    * Variable substitution is performed on <code>value</code> before it is
    * added to the map.
    *
    * @param   key
    *          A name which uniquely identifies the parameter.
    * @param   value
    *          The parameter's value.
    *
    * @return  The text of <code>value</code> after variable substitution
    *          has been performed.
    *
    * @throws  ConfigurationException
    *          if <code>key</code> already identifies a configuration
    *          parameter.
    */
   public static String putParameter(String key, String value)
   throws ConfigurationException
   {
      if (key == null || value == null)
      {
         throw new ConfigurationException("Null parameter key or value");
      }
      
      if (parmMap.containsKey(key))
      {
         throw new ConfigurationException("Duplicate parameter: " + key);
      }
      
      // handle variable substitution
      value = substituteParms(value);
      
      parmMap.put(key, value);
      
      return value;
   }
   
   /**
    * Variable substitution is performed on <code>value</code> before it is
    * returned.  Any portions of the string matching a ${SYMBOL} syntax will
    * have the parameter map searched for SYMBOL. If it is found, the value
    * of that parameter will be inserted in place of ${SYMBOL}.
    * <p>
    * Replacements occur in the same order they were defined in the
    * main configuration file, and then in the order in which they were
    * added by external callers of this method.
    *
    * @param   value
    *          The value on which to operate.
    *
    * @return  The text of <code>value</code> after variable substitution
    *          has been performed.
    */
   public static String substituteParms(String value)
   {
      return StringHelper.sweep(value, parmMap);
   }
   
   /**
    * Get the P2J home path defined for the application or "." if no such
    * path is defined.
    *
    * @return   P2J home path or "." if the system property has not been 
    *           defined.
    *
    * @throws   ConfigurationException
    *           If there is any problem with normalizing the specified path.
    */
   public static String home()
   throws ConfigurationException
   {
      // Check for the P2J_HOME system property.
      String home = System.getProperty(PROP_HOME_DIR, ".");
      
      // Normalize and verify the path.
      File dir = new File(home);
      home = normalizePath(dir, true, true);
      
      return home;
   }
   
   /**
    * If <code>path</code> represents a relative path, generate a
    * <code>File</code> object for this path, relative to the home directory
    * defined for the application. If <code>filename</code> is an absolute
    * path, simply use it to construct the <code>File</code>.
    *
    * @param   path
    *          A relative or absolute file or directory path.
    *
    * @throws  ConfigurationException
    *          if <code>path</code> is relative, but no home path has been
    *          defined for the application.
    */
   public static File toFile(String path)
   throws ConfigurationException
   {
      File file = new File(path);
      if (!file.isAbsolute())
      {
         file = new File(home(), path);
         try
         {
            file = file.getCanonicalFile();
         }
         catch (IOException exc)
         {
         }
      }
      
      return file;
   }
   
   /**
    * Converts an absolute or relative filename (relative to the current
    * directory) into a filename that is relative to the project home. This 
    * allows tools referencing the same file from different current
    * directories to reference the same file, even though the pathing may 
    * not appear to be an exact match.
    *
    * @param    filename
    *           Absolute or current directory relative filename.
    *
    * @return   Filename normalized to the project root OR the original
    *           filename if the file exists outside of the project directory
    *           tree (and thus cannot be normalized).
    */
   public static String normalizeFilename(String filename)
   {
      File   abnormal = new File(filename); 
      String absolute;
      String home;
      
      try
      {
         home = home();
      }
      catch (ConfigurationException cfg)
      {
         // we are done, there is no normalization to be done if we can't
         // find the project root
         return filename;
      }
      
      try
      {
         // make best effort to clean up the path
         absolute = abnormal.getCanonicalPath();
      }
      catch (IOException exc)
      {
         // fallback plan if cannonical doesn't work
         absolute = abnormal.getAbsolutePath();
      }
      
      // make sure this filename is in the project directory tree
      if (absolute.startsWith(home))
      {
         // all is well, we can normalize this path
         StringBuilder sb = new StringBuilder(".");
         sb.append(absolute.substring(home.length()));
         
         return sb.toString();
      }
      
      // the path can't be normalized (because it exists outside of the
      // project directory tree, so we can only return the original filename
      return filename;
   }   
   
   /**
    * Get the configuration directory, which must be a <code>cfg</code>
    * subdirectory beneath the P2J home path.
    *
    * @return  Configuration subdirectory.
    *
    * @throws  ConfigurationException
    *          if configuration directory does not exist or is not a
    *          directory.
    */
   private static File getConfigDirectory()
   throws ConfigurationException
   {
      File dir = new File(home(), CFG_DIR);
      normalizePath(dir, true, true);
      
      return dir;
   }
   
   /**
    * Get the configuration file for the subsystem associated with
    * <code>root</code>, which is either <code>&lt;root&gt;.cfg.xml</code>
    * or, if this file is missing, is the default configuration file,
    * <code>p2j.cfg.xml</code>.
    *
    * @param   dir
    *          Configuration directory.
    * @param   root
    *          Name associated with the target subsystem.
    *
    * @return  Configuration file.
    *
    * @throws  ConfigurationException
    *          if configuration file cannot be found.
    */
   private static File getConfigFile(File dir, String root)
   throws ConfigurationException
   {
      File file = new File(dir, root + CFG_EXTENSION);
      
      if (!file.exists())
      {
         file = new File(dir, DEF_CFG_FILE);
      }
      
      if (!file.exists())
      {
         throw new ConfigurationException(
            "No data file found for " + root + " configuration");
      }
      
      return file;
   }
   
   /**
    * Normalize and optionally verify the specified file system object.
    * <p>
    * Normalizing the path converts relative paths to absolute paths and
    * attempts to remove any redundant, relative constructs, such as
    * references to the current directory (<code>.</code>) and to the parent
    * directory (<code>..</code>).
    * <p>
    * Verifying the path checks that the object exists in the file system
    * and that it is the expected type (file vs. directory).
    *
    * @param   file
    *          File whose path is to be manipulated.
    * @param   verify
    *          <code>true</code> to verify the path, else <code>false</code>.
    * @param   isDirectory
    *          <code>true</code> if <code>file</code> is supposed to represent
    *          a directory, <code>false</code> if it should represent a file.
    *          This argument is only relevant if <code>verify</code> is set to
    *          <code>true</code>.
    *
    * @throws  ConfigurationException
    *          if verification fails.
    */
   private static String normalizePath(File file,
                                       boolean verify,
                                       boolean isDirectory)
   throws ConfigurationException
   {
      String path = null;
      
      try
      {
         path = file.getCanonicalPath();
      }
      catch (IOException exc)
      {
         path = file.getAbsolutePath();
      }
      
      if (verify)
      {
         if (!file.exists())
         {
            throw new ConfigurationException("'" + path + "' does not exist");
         }
         
         if (file.isDirectory() != isDirectory)
         {
            String type = (isDirectory ? "directory" : "file");
            throw new ConfigurationException(
               "'" + path + "' is not a " + type);
         }
      }
      
      return path;
   }
   
   /**
    * Prepare the general purpose SAX handler for working with P2J XML config 
    * file syntax.  
    * 
    * @param saxHandler
    *        The SAX handler to be configured.
    */
   private static void setupHandler(GenericSAXHandler saxHandler)
   {
      // XML/object mappers:
      saxHandler.installMapper(TAG_PARAMETER, 
                               new Mapper<Parameter>(Parameter.class));
      
      saxHandler.installMapper(TAG_SCHEMA,
                               new Mapper<SchemaConfig>(SchemaConfig.class)
         {
            @Override
            public void mergeSubObj(SchemaConfig s, Object c)
            {
               checkSubObjTypes(c,
                  new Class<?>[] {NsConfig.class, Metadata.class});
               if (c instanceof NsConfig)
               {
                  s.addNamespace((NsConfig)c);
               }
               else
               {
                  s.setMetadata((Metadata) c);
               }
            }
         });
      
      saxHandler.installMapper(TAG_NAMESPACE, 
         new Mapper<NsConfig>(NsConfig.class)
         {
            @Override
            public void mergeSubObj(NsConfig ns, Object c)
            {
               checkSubObjTypes(c, 
                  new Class<?>[] {Parameter.class, DialectParam.class});
               if (c instanceof Parameter)
               {
                  ns.addParameter((Parameter)c);
               }
               else
               {
                  ns.addDialectParameter((DialectParam)c);
               }
            }
         });
  
      saxHandler.installMapper(TAG_DIALECTSPECIFIC, 
                               new Mapper<DialectParam>(DialectParam.class)
         {
            @Override
            public void mergeSubObj(DialectParam d, Object c)
            {
               checkSubObjType(c, Parameter.class);
               d.addParameter((Parameter)c);
            }
         });
      
      // throwaway class used just to map the <table> tag so that we can
      // add a table name to the Metadata object in the Mapper below
      final class Table
      {
         private String name = null;
         
         public String getName()
         {
            return name;
         }
         
         public void setName(String name)
         {
            this.name = name;
         }
      }
      
      saxHandler.installMapper(TAG_METADATA,
                               new Mapper<Metadata>(Metadata.class)
         {
            @Override
            public void mergeSubObj(Metadata m, Object c)
            {
               checkSubObjType(c, Table.class);
               m.addTable(((Table) c).getName());
            }
         });
      
      saxHandler.installMapper(
               TAG_TABLE,
               new Mapper<Table>(Table.class));
      
      // tags that don't map to objects:
      saxHandler.setUnhandledTags(new String[] {TAG_CFG, TAG_GLOBAL});

      // constraints:
      saxHandler.setConstraints(TAG_PARAMETER, 
                                new String[]{"name", "value"}, 
                                null, 
                                new String[]{}, 
                                new String[]{});
      
      saxHandler.setConstraints(TAG_NAMESPACE, 
         new String[]{"name", "importFile", "xmlFile"}, 
         new String[]{"default"}, 
         null, 
         new String[]{
                      TAG_PARAMETER, 
                      TAG_DIALECTSPECIFIC
                     });
      
      saxHandler.setConstraints(TAG_DIALECTSPECIFIC, 
         new String[]{"name"}, 
         new String[]{}, 
         new String[]{TAG_PARAMETER}, 
         new String[]{});
      
      saxHandler.setConstraints(TAG_METADATA,
         new String[]{"name"},
         null,
         null,
         new String[]{"table"});
      
      saxHandler.setConstraints(TAG_TABLE,
         new String[]{"name"},
         null,
         null,
         null);
   }
   
   /**
    * Parses a given XML file using the SAX parser and handler that were 
    * already prepared. The objects built in the process are held in the 
    * handler and can be fetched by XML tag name with {@link 
    * GenericSAXHandler#getGlobalObjects(String)}.  
    * 
    * @param  cfgFile
    *         The file to be parsed.   
    *         
    * @throws IOException 
    *         If any IO errors occur in reading the file.
    * @throws SAXException 
    *         If any SAX errors occur during processing.
    * @throws RuntimeException
    *         wrapping a <code>ConfigurationException</code> if any error 
    *         occurs reading the XML data or creating the configuration 
    *         object. This includes missing configuration file or file format
    *         constraint violations detected during parsing.
    *              
    * @see    GenericSAXHandler#getGlobalObjects(String)
    */
   private static void loadXml(File cfgFile)
   throws SAXException, 
          IOException,
          RuntimeException
   {
      FileInputStream fis = new FileInputStream(cfgFile);
      saxParser.reset();
      saxHandler.resetState();
      saxParser.parse(new InputSource(new FileReader(cfgFile)), saxHandler);
      fis.close();
   }
   
   /**
    * Simple schema loader test: the XML file passed as an argument is parsed
    * and the resulting <code>SchemaConfig</code> together with all global
    * parameters are printed to standard output. Variable substitution is not
    * performed on parameters, their values are shown as they were read from 
    * XML. 
    * 
    * @param  args
    *         The first element is the absolute path of the XML file to parse.
    *         
    * @throws IOException
    *         If IO errors occur. 
    * @throws SAXException
    *         If the SAX parser encounters problems while processing the XML
    *         content.
    * @throws RuntimeException
    *         wrapping a <code>ConfigurationException</code> if any error 
    *         occurs reading the XML data or creating the configuration 
    *         objects. This includes missing configuration file or file format
    *         constraint violations detected during parsing.
    */
   public static void main(String[] args) 
   throws SAXException, 
          IOException,
          RuntimeException
   {
      //parse the file
      loadXml(new File(args[0]));
      
      //print the constructed objects:
      
      SchemaConfig sc = (SchemaConfig) saxHandler.getGlobalObjects(
         TAG_SCHEMA).get(0);
      System.out.println("Global schema object:\n\n" + sc);
      
      System.out.println("\nGlobal parameters:\n");
      List<?> params = saxHandler.getGlobalObjects(TAG_PARAMETER);
      for (Object o : params)
      {
         Parameter p = (Parameter)o;
         System.out.println("name = " + p.getName() + ", value = " + 
            p.getValue());
      }
   }
}
